uses Music, Crt, MsGraph;

const
	FieldWidth = 105;
	FieldHeight = 77;
	MoveKeys : array [0..3, 0..1] of string [2] =
		((#0#75, #0#77), ('Z', 'X'), (',', '.'), ('2', '3'));

type
	mvdir = (North, East, South, West);
	Beam = record
		headX : byte;
		headY : byte;
		direction : mvdir;
		alive : boolean
	end;

{$I BeamHead.inc}
{$I BeamBody.inc}
{$I BeamStrt.inc}
{$I BeamCol.inc}

var
	NoOfPlayers, NoOfRounds, Speed : byte;
	Border, PlaySound : boolean;
	winner : byte;
	endgame : boolean;
	a, i : byte;
	Score : array [0..4] of byte;
	Draws : byte;
	c, junk : char;

procedure Palette (index : byte; value : longint);
{ For faster output }
begin
	port [$3c8] := index;
	port [$3c9] := Byte (value);
	port [$3c9] := Byte (value shr 8);
	port [$3c9] := Byte (value shr 16);
end;

function MidColor (C1, C2 : longint; steps, curstep : byte) : longint;
begin
	MidColor := (C1 shl 8 shr 24 * (steps - curstep) +
							 C2 shl 8 shr 24 * curstep) div steps shl 16 +
							(C1 shl 16 shr 24 * (steps - curstep) +
							 C2 shl 16 shr 24 * curstep) div steps shl 8 +
							(C1 shl 24 shr 24 * (steps - curstep) +
							 C2 shl 24 shr 24 * curstep) div steps
end;

procedure DrawBeamHead (player, X, Y : byte; dir : mvdir);
begin
	_PutImage (6*X+5, 6*Y+17, BeamHead [player, dir], _GPSet);
end;

{procedure ClearBeamHead (X, Y : byte; dir : mvdir);
begin
end;}

procedure DrawBeamBody (player, X, Y : byte; dir1, dir2 : mvdir);
begin
	dir1 := MvDir (byte (dir1) xor 2);
	if (dir1 = North) or (dir2 = North) then Dec (dir1);
	_PutImage (6*X+5, 6*Y+17, BeamBody [player, byte (byte (dir1) + byte (dir2))], _GPSet);
end;

procedure Fade (player : byte; var FinishedBeams : byte);
const CurFadeStep : array [0..3] of byte = (0, 0, 0, 0);
			FadeSteps = 32;
var i, j, a : byte;
begin
	if player = 255 then
	begin
		FinishedBeams := 0;
		FillChar (CurFadeStep, SizeOf (CurFadeStep), 0);
		for j := 0 to 3 do
			for i := 0 to 2 do
				Palette (j * 3 + i + 1, AliveBeamColor [j, i]);
		for i := 0 to 2 do
			Palette (i + 13, BorderColor [i]);
	end
	else if CurFadeStep [player] <> FadeSteps then
	begin
		Inc (CurFadeStep [player]);
		for i := 0 to 2 do
			Palette (player*3+i+1, MidColor (AliveBeamColor [player, i],
		DeadBeamColor [player, i], FadeSteps, CurFadeStep [player]));
		if CurFadeStep [player] = FadeSteps then Inc (FinishedBeams);
	end
end;

procedure DieMusic;
begin
	MusicOff;
	PlayMusic ('MLL64O0C>C>C>C>C>C>C<C<C<C<C<C')
end;

procedure MoveMusic;
begin
	if not MusicBusy then PlayMusic ('O-2C256');
end;

procedure WaitTime;
var Timer : longint absolute $40:$6C;
		OldTimer : longint;
begin
	OldTimer := Timer;
	while Timer = OldTimer do;
end;

function Session (NoOfPlayers, RoundNo, Speed : byte;
									Border, PlaySound : boolean;
									var endgame : boolean) : byte; {Winner}

const MvBufSize = 16;
var Board : array [0..FieldWidth-1, 0..FieldHeight-1] of boolean;
		Beams : array [0..3] of Beam;
		MoveBuf : array [0..3, 0..MvBufSize-1] of mvdir;
		BufReadPos, BufWritePos : array [0..MvBufSize-1] of byte;
		NoOfMoves : Word;
		NewDir : mvdir;
		AliveBeams, FinishedBeams, Tick : byte;
		i : byte;
		j : integer;

	procedure ReadMvBuf;
	var KeyBuf : string [2];
			i : byte;
	begin
		while KeyPressed do
		begin
			KeyBuf := ReadKey;
			if KeyBuf = #0 then KeyBuf := #0 + ReadKey
				else if (KeyBuf >= 'a') and (KeyBuf <= 'z') then
					KeyBuf := Chr (Ord (KeyBuf [1]) - 32);
			i := 0;
			while (i <= 7) and (KeyBuf <> MoveKeys [i shr 1, i and 1]) do Inc (i);
			if i <= 7 then
			begin
				if i and 1 = 0 then
					MoveBuf [i shr 1, BufWritePos [i shr 1]] :=
					MvDir ((byte (MoveBuf [i shr 1, (BufWritePos [i shr 1]
					+ MvBufSize - 1) mod MvBufSize]) + 3) mod 4)
				else
					MoveBuf [i shr 1, BufWritePos [i shr 1]] :=
					MvDir ((byte (MoveBuf [i shr 1, (BufWritePos [i shr 1]
					+ MvBufSize - 1) mod MvBufSize]) + 1) mod 4);
				BufWritePos [i shr 1] := (BufWritePos [i shr 1] + 1) mod MvBufSize;
			end
			else
				if KeyBuf = #27 then endgame := True;
		end;
	end;

	procedure ProcessMvBuf (Player : byte; var NewDir : mvdir);
	begin
		if BufReadPos [Player] <> BufWritePos [Player] then
		begin
			NewDir := MoveBuf [Player, BufReadPos [Player]];
			BufReadPos [Player] := (BufReadPos [Player] + 1) mod MvBufSize;
		end;
	end;

	procedure InitField;
	var i : byte;
	begin
		_ClearScreen (_GClearScreen);
		AliveBeams := NoOfPlayers;
		FillChar (Board, SizeOf (Board), 0);
		Fade (255, FinishedBeams);
		if Border then
		begin
			for i := 0 to FieldWidth - 1 do
			begin
				Board [i, 0] := True;
				Board [i, FieldHeight - 1] := True;
			end;
			for i := 0 to FieldHeight - 1 do
			begin
				Board [0, i] := True;
				Board [FieldWidth - 1, i] := True;
			end;
			_SetColor (14);
			_Rectangle (_GFillInterior, 4, 16, 635, 479);
			_SetColor (0);
			_Rectangle (_GFillInterior, 11, 23, 628, 472);
			_SetColor (15);
			_Rectangle (_GFillInterior, 4, 16, 634, 16);
			_Rectangle (_GFillInterior, 4, 16, 4, 478);
			_Rectangle (_GFillInterior, 629, 23, 629, 473);
			_Rectangle (_GFillInterior, 11, 473, 629, 473);
			_SetColor (13);
			_Rectangle (_GFillInterior, 635, 17, 635, 479);
			_Rectangle (_GFillInterior, 5, 479, 635, 479);
			_Rectangle (_GFillInterior, 10, 22, 628, 22);
			_Rectangle (_GFillInterior, 10, 22, 10, 472);
		end;
		NoOfMoves := 0;
		endgame := False;
		FillChar (BufReadPos [0], SizeOf (BufReadPos), 0);
		FillChar (BufWritePos [0], SizeOf (BufWritePos), 0);
		Move (BeamStart [NoOfPlayers * (NoOfPlayers - 1) div 2],
		Beams, SizeOf (Beams [0]) * NoOfPlayers);
		TextAttr := $f;
		GotoXY (2, 1);
		Write ('Round ', RoundNo);
		GotoXY (14, 1);
		Write ('Moves : ');
		for i := 0 to NoOfPlayers - 1 do
		begin
			Beams [i].direction := MvDir (Random (4));
			MoveBuf [i, MvBufSize - 1] := Beams [i].direction;
			Board [Beams [i].headX, Beams [i].headY] := True;
			DrawBeamHead (i, Beams [i].headX, Beams [i].headY, Beams [i].direction);
			if NoOfPlayers > 1 then
			begin
				TextAttr := 3 + 3 * i;
				case i of
					0 : begin GotoXY (29, 1); Write ('Red : ') end;
					1 : begin GotoXY (41, 1); Write ('Green : ') end;
					2 : begin GotoXY (55, 1); Write ('Blue : ') end;
					3 : begin GotoXY (68, 1); Write ('Yellow : ') end;
				end;
				Write (score [i] : 3);
			end;
		end;
	end;

begin

	InitField;
	Tick := 0;
	repeat
		WaitTime;
		Inc (Tick);
		ReadMvBuf;
		if (Tick = Speed) and ((AliveBeams > 1) and (NoOfPlayers > 1)
												or (AliveBeams > 0) and (NoOfPlayers = 1)) then
		begin
			Inc (NoOfMoves);
			TextAttr := 15;
			GotoXY (22, 1);
			Write (NoOfMoves : 4);
			Tick:=0;
			for i := 0 to NoOfPlayers - 1 do
				with Beams [i] do
					if alive then
					begin
						NewDir := direction;
						ProcessMvBuf (i, NewDir);
						j := 0;
						while (headX <> Beams [j].headX) or (headY <> Beams [j].headY)
							do Inc (j);
						if j = i then DrawBeamBody (i, headX, headY, direction, NewDir);
						direction := NewDir;
						case NewDir of
							North : headY := (headY + FieldHeight - 1) mod FieldHeight;
							East  : headX := (headX + 1) mod FieldWidth;
							South : headY := (headY + 1) mod FieldHeight;
							West  : headX := (headX + FieldWidth - 1) mod FieldWidth;
						end;
						if Board [headX, headY] then alive := False
							else Board [headX, headY] := True;
						if not alive then
						begin
{			 	      ClearBeamHead (headX, headY, NewDir);}
							if PlaySound then DieMusic;
							Dec (AliveBeams)
						end;
						DrawBeamHead (i, headX, headY, NewDir);
						if not alive then
						begin
							for j := 0 to i - 1 do
							if (headX = Beams [j].headX) and (headY = Beams [j].headY) then
							begin
								if Beams [j].alive then Dec (AliveBeams);
								Beams [j].alive := False;
								DrawBeamHead (i, headX, headY, NewDir);
							end;
						end
					end;
			if PlaySound then MoveMusic;
		end;
		for i := 0 to NoOfPlayers - 1 do
			if not Beams [i].alive then
				Fade (i, FinishedBeams);
	until endgame or ((FinishedBeams >= NoOfPlayers - 1) and (NoOfPlayers > 1)
								 or (FinishedBeams >= NoOfPlayers) and (NoOfPlayers = 1));
	Session := 255;
	if AliveBeams = 1 then
	begin
		i := 0;
		while not (Beams [i].alive) do Inc(i);
		Session := i;
	end;
end;

procedure GetInputs (var NoOfPlayers, NoOfRounds, Speed : byte;
										 var Border, PlaySound : boolean);
var a, junk : char;
		i, j : integer;
		s : string;
begin

	i := _SetVideoMode (_TextC40);
	ClrScr;
	TextAttr := $f;
	GotoXY (17, WhereY);
	WriteLn ('B E A M'#13#10);

	TextAttr := 3;
	WriteLn ('Beam is a game for two to four players.');
	WriteLn ('Navigate your beam avoiding running into');
	WriteLn ('beams or border (including yourself).');
	WriteLn ('At the same time, make others crash into');
	WriteLn ('something.  The one who survives wins.');
	WriteLn;

	TextAttr :=  7; Write ('How many players ? <');
	TextAttr := $f; Write ('2');
	TextAttr :=  7; Write ('/');
	TextAttr := $f; Write ('3');
	TextAttr :=  7; Write ('/');
	TextAttr := $f; Write ('4');
	TextAttr :=  7; Write ('/');
	TextAttr := $f; Write ('P');
	TextAttr :=  7; Write ('ractise> ');
	repeat
		a := ReadKey;
		if a = #0 then junk := ReadKey;
		a := UpCase (a);
	until a in ['2', '3', '4', 'P'];
	WriteLn (a);
	WriteLn;
	if a = 'P' then
		NoOfPlayers := 1
	else
		NoOfPlayers := Ord (a) - 48;

	WriteLn ('How many rounds ? (1 - 100)');
	repeat
		GotoXY (29, 11);
		Write ('' : 127);
		GotoXY (29, 11);
		ReadLn (s);
		Val (s, i, j);
	until (i >= 1) and (i <= 100) and (Length (s) <= 3);
	WriteLn;
	NoOfRounds := i;

	Write ('Speed ? <');
	TextAttr := $f; Write ('N');
	TextAttr :=  7; Write ('ormal/');
	TextAttr := $f; Write ('S');
	TextAttr :=  7; Write ('low> ');
	repeat
		a := ReadKey;
		if a = #0 then junk := ReadKey;
		a := UpCase (a);
	until a in ['N', 'S'];
	WriteLn (a);
	if a = 'N' then
		Speed := 1
	else
		Speed := 2;

	Write ('Are borders present ? <');
	TextAttr := $f; Write ('Y');
	TextAttr :=  7; Write ('/');
	TextAttr := $f; Write ('N');
	TextAttr :=  7; Write ('> ');
	repeat
		a := ReadKey;
		if a = #0 then junk := ReadKey;
		a := UpCase (a);
	until a in ['Y', 'N'];
	WriteLn (a);
	if a = 'Y' then
		Border := True
	else
		Border := False;

	Write ('Do you want to hear sound ? <');
	TextAttr := $f; Write ('Y');
	TextAttr :=  7; Write ('/');
	TextAttr := $f; Write ('N');
	TextAttr :=  7; Write ('> ');
	repeat
		a := ReadKey;
		if a = #0 then junk := ReadKey;
		a := UpCase (a);
	until a in ['Y', 'N'];
	WriteLn (a);
	if a = 'Y' then
		PlaySound := True
	else
		PlaySound := False;

	TextAttr := $f;
	GotoXY (14, WhereY + 1);
	Write ('Game Controls');
	TextAttr := 7;
	GotoXY (20, WhereY + 2);
	WriteLn ('Left Right');

	for i := 0 to NoOfPlayers - 1 do
	begin
		GotoXY (11, WhereY);
		case i of
			0 : begin TextAttr := $c; WriteLn ('Red    :   '#27'    '#26) end;
			1 : begin TextAttr := $a; WriteLn ('Green  :   Z    X') end;
			2 : begin TextAttr := $9; WriteLn ('Blue   :   ,    .') end;
			3 : begin TextAttr := $e; WriteLn ('Yellow :   2    3') end;
		end;
	end;

	TextAttr := $7;
	GotoXY (11, 25);
	Write ('Press space to play');
	while ReadKey <> ' ' do
		if KeyPressed then junk := ReadKey;

end;

procedure PrintResult;
var i : byte;
begin
	TextAttr := $f;
	if NoOfPlayers > 1 then
	begin
		GotoXY (15, 11 - NoOfPlayers shr 1);
		Write ('Final Scores');
		for i := 0 to NoOfPlayers - 1 do
		begin
			GotoXY (15, 13 + i - NoOfPlayers shr 1);
			case i of
				0 : begin TextAttr := $c; Write ('Red    : ') end;
				1 : begin TextAttr := $a; Write ('Green  : ') end;
				2 : begin TextAttr := $9; Write ('Blue   : ') end;
				3 : begin TextAttr := $e; Write ('Yellow : ') end;
			end;
			Write (score [i] : 3);
		end;
		TextAttr := $f;
		GotoXY (15, 13 + (NoOfPlayers + 1) shr 1);
		Write ('Draw   : ', Draws : 3);
	end
	else
	begin
		GotoXY (10, 10);
		Write ('The practise was over!');
	end;
end;

begin
	CheckBreak := False;
	Randomize;
	repeat
		GetInputs (NoOfPlayers, NoOfRounds, Speed, Border, PlaySound);
		FillChar (Score, SizeOf (Score), 0);
		Draws := 0;
		DirectVideo := False;
		a := _SetVideoMode (_VRes16Color);
		endgame := False;
		i := 0;
		repeat
			Inc (i);
			winner := Session (NoOfPlayers, i, Speed, Border, PlaySound, endgame);
			if winner = 255 then
				Inc (Draws)
			else
				Inc (Score [winner]);
		until (i = NoOfRounds) or endgame;
		i := _SetVideoMode (_TextC40);
		DirectVideo := True;
		PrintResult;
		GotoXY (4, 15 + (NoOfPlayers + 1) shr 1);
		Write ('Do you like to play again ? <Y/N> ');
		repeat
			c := ReadKey;
			if c = #0 then junk := ReadKey;
			c := UpCase (c);
		until c in ['Y', 'N'];
	until c = 'N';
	a := _SetVideoMode (_DefaultMode)

end.
