unit music;
Interface
	procedure MusicOff;
	procedure PlayMusic(s : string);
	procedure PlayMusicForeground(s : string);
	function MusicBusy : boolean;

Implementation

uses dos;

var
	OldInt1C        : pointer;
	ExitSave        : pointer;
	MusicString     : string;
	MusicHere       : integer;
	MusicDelay1     : integer;
	MusicDelay2     : integer;
	MusicNoteLength : integer;
	MusicTempo      : integer;
	MusicOctave     : integer;
	MusicKind       : integer;
	Frequency       : array[-30..113] of integer;

function GetNumber(min, max, default : integer) : integer;
var
	n : integer;
	sign : integer;
begin
if (MusicHere <= length(MusicString)) and (MusicString[MusicHere] = '=') then
	begin
	while (MusicHere <= length(MusicString)) and (MusicString[MusicHere] <> ';')
		do inc(MusicHere);
	if (MusicHere <= length(MusicString)) and (MusicString[MusicHere] = ';')
		then inc(MusicHere);
	GetNumber := default;
	exit;
	end;
n := 0;
sign := 1;
if (MusicHere <= length(MusicString)) and
	(MusicString[MusicHere] = '-') then
begin
	sign := -1;
	inc(MusicHere);
end;
while (MusicHere <= length(MusicString)) and
	(MusicString[MusicHere] in ['0'..'9']) do
	begin
	n := n * 10 + (Ord(MusicString[MusicHere]) - Ord('0'));
	inc(MusicHere);
	end;
if (n < min) or (n > max)
	then GetNumber := default
	else GetNumber := n * sign;
end;

procedure SetupDelays;
var
	r : real;
begin
r := GetNumber(1,256,MusicNoteLength);
while (MusicHere <= length(MusicString)) and
	 (MusicString[MusicHere] = '.') do
	begin
	inc(MusicHere);
	r := r * 0.75;
	end;
MusicDelay1 := Round(4368.0 / (r * MusicTempo));
if MusicKind < 8
	then MusicDelay2 := MusicDelay1 * (8 - MusicKind) div 8
	else MusicDelay2 := 0;
dec(MusicDelay1,MusicDelay2);
end;

procedure MusicNext; interrupt;
var
	note : integer;
	ch : char;
begin
Inline(
	$9C/                   {pushf}
	$FF/$1E/>OLDINT1C);    {call far [>OldInt1C]}
if MusicDelay1 > 0 then
	begin
	dec(MusicDelay1);
	if MusicDelay1 > 0 then exit;
	end;
if MusicDelay2 > 0 then
	begin
	MusicDelay1 := MusicDelay2;
	MusicDelay2 := 0;
	Port[$61] := Port[$61] and $F8;
	exit;
	end;
if MusicHere = 0 then	exit;
if MusicHere > length(MusicString) then
	begin
	MusicHere := 0;
	Port[$61] := Port[$61] and $F8;
	exit;
	end;
while MusicHere <= length(MusicString) do
	begin
	ch := upcase(MusicString[MusicHere]);
	inc(MusicHere);
	case ch of
		'O' : MusicOctave := GetNumber(-2,9,4);
		'L' : MusicNoteLength := GetNumber(1,256,4);
		'T' : MusicTempo := Getnumber(32,255,120);
		'M' : if MusicHere <= length(MusicString) then
					begin
					ch := upcase(MusicString[MusicHere]);
					inc(MusicHere);
					case ch of
						'L' : MusicKind := 8;
						'N' : MusicKind := 7;
						'S' : MusicKind := 6;
						end;
					end;
		'P' : begin
					Port[$61] := Port[$61] and $F8;
					SetupDelays;
					exit;
					end;
		'A'..'G','>','<' : begin
					note := MusicOctave * 12;
					if ch = '>' then
						begin
						if MusicHere <= length(MusicString) then
							ch := upcase(MusicString[MusicHere]);
						inc(MusicHere);
						if note <= 71 then
							begin
								inc(note,12);
								inc(MusicOctave);
							end;
						end;
					if ch = '<' then
						begin
						if MusicHere <= length(MusicString) then
							ch := upcase(MusicString[MusicHere]);
						inc(MusicHere);
						if note >= 12 then
							begin
								dec(note,12);
								dec(MusicOctave);
							end;
						end;
					case ch of
						'A' : inc(note,9);
						'B' : inc(note,11);
						'C' : inc(note,0);
						'D' : inc(note,2);
						'E' : inc(note,4);
						'F' : inc(note,5);
						'G' : inc(note,7);
						end;
					if (MusicHere <= length(MusicString)) and
						 ( (MusicString[MusicHere] = '#') or (MusicString[MusicHere] = '+') )
						 then
						begin
						inc(MusicHere);
						inc(note);
						end;
					if (MusicHere <= length(MusicString)) and
						 (MusicString[MusicHere] = '-') then
						begin
						inc(MusicHere);
						dec(note);
						end;
					note := Frequency[note];
					Port[$61] := Port[$61] and $F8;
					Port[$43] := $B6;
					Port[$42] := Lo(note);
					Port[$42] := Hi(note);
					Port[$61] := Port[$61] or $03;
					SetupDelays;
					exit;
					end;
		'N' : begin
					note := GetNumber(-30,113,0);
					Port[$61] := Port[$61] and $F8;
					if note > 0 then
						begin
						note := Frequency[note-1];
						Port[$43] := $B6;
						Port[$42] := Lo(note);
						Port[$42] := Hi(note);
						Port[$61] := Port[$61] or $03;
						end;
					SetupDelays;
					exit;
					end;
		'X' : begin
					while (MusicHere <= length(MusicString)) and
						(MusicString[MusicHere] <> ';') do inc(MusicHere);
					if (MusicHere <= length(MusicString)) and
						(MusicString[MusicHere] = ';') then inc(MusicHere);
					end;
		end;
	end;
end;

procedure MusicOff;
{ Turn music off. }
begin
MusicHere := 0;
MusicDelay1 := 0;
MusicDelay2 := 0;
Port[$61] := Port[$61] and $F8;
end;

procedure PlayMusic(s : string);
begin
MusicOff;
MusicString := s;
MusicNoteLength := 4;
MusicTempo := 120;
MusicOctave := 4;
MusicKind := 7;
MusicHere := 1;
end;

procedure PlayMusicForeground(s : string);
begin
PlayMusic(s);
while MusicHere > 0 do ;
end;

function MusicBusy : boolean;
begin
if MusicHere > 0
	then MusicBusy := true
	else MusicBusy := false;
end;

{$F+}

procedure ShutDown;
begin
MusicOff;
ExitProc := ExitSave;
SetIntVec($1C,OldInt1C);
end;

{$F-}

procedure Initialize;
var
	i : integer;
	r1, r2 : real;
begin
r1 := 1193180.0 / 8000.0 /8.0;
r2 := exp(ln(2.0)/12.0);
for i := 113 downto -30 do
	begin
	Frequency[i] := round(r1);
	r1 := r1 * r2;
	end;
MusicOff;
GetIntVec($1C,OldInt1C);
SetIntVec($1C,@MusicNext);
ExitSave := ExitProc;
ExitProc := @ShutDown;
end;

begin
	Initialize;
end.
